@@ -1,4 +1,4 @@
-/* g10.c - The G10 re-install utility
+/* g10.c - The G10 utility
  *	Copyright (c) 1997 by Werner Koch (dd9jn)
  *
  * This file is part of G10.
@@ -35,6 +35,15 @@
 #include "cipher.h"
 #include "filter.h"
 
+enum cmd_values { aNull = 0,
+    aSym, aStore, aEncr, aPrimegen, aKeygen, aSign, aSignEncr,
+    aPrintMDs, aSignKey, aClearsig, aListPackets, aEditSig,
+    aKMode, aKModeC,
+aTest };
+
+
+static void set_cmd( enum cmd_values *ret_cmd,
+			enum cmd_values new_cmd );
 static void print_hex( byte *p, size_t n );
 static void print_mds( const char *fname );
 static void do_test(int);
@@ -94,6 +103,27 @@ set_debug(void)
 }
 
 
+static void
+set_cmd( enum cmd_values *ret_cmd, enum cmd_values new_cmd )
+{
+    enum cmd_values cmd = *ret_cmd;
+
+    if( !cmd || cmd == new_cmd )
+	cmd = new_cmd;
+    else if( cmd == aSign && new_cmd == aEncr )
+	cmd = aSignEncr;
+    else if( cmd == aEncr && new_cmd == aSign )
+	cmd = aSignEncr;
+    else if( cmd == aKMode && new_cmd == aSym )
+	cmd = aKModeC;
+    else {
+	log_error("conflicting commands\n");
+	exit(2);
+    }
+
+    *ret_cmd = cmd;
+}
+
 
 int
 main( int argc, char **argv )
@@ -116,12 +146,13 @@ main( int argc, char **argv )
     { 508, "check-key" ,0, "check signatures on a key in the keyring" },
     { 509, "keyring"   ,2, "add this keyring to the list of keyrings" },
     { 's', "sign",      0, "make a signature"},
+    { 't', "textmode",  0, "use canonical text mode"},
     { 'b', "detach-sign", 0, "make a detached signature"},
     { 'e', "encrypt",   0, "encrypt data" },
     { 'd', "decrypt",   0, "decrypt data (default)" },
-  /*{ 'c', "check",     0, "check a signature (default)" }, */
-    { 'l', "local-user",2, "use this user-id to sign or decrypt" },
+    { 'u', "local-user",2, "use this user-id to sign or decrypt" },
     { 'r', "remote-user", 2, "use this user-id for encryption" },
+    { 'k', NULL      , 0, "list keys" },
     { 510, "debug"     ,4|16, "set debugging flags" },
     { 511, "debug-all" ,0, "enable full debugging"},
     { 512, "cache-all" ,0, "hold everything in memory"},
@@ -134,14 +165,15 @@ main( int argc, char **argv )
     { 518, "options"   , 2, "read options from file" },
     { 519, "no-armor",   0, "\r"},
     { 520, "no-default-keyring", 0, "\r" },
+    { 521, "list-packets",0,"list only the sequence of packets"},
+    { 522, "no-greeting", 0, "\r" },
+    { 523, "passphrase-fd",1, "\r" },
+    { 524, "edit-sig"  ,0, "edit a key signature" },
 
     {0} };
     ARGPARSE_ARGS pargs;
     IOBUF a;
     int rc;
-    enum { aNull, aSym, aStore, aEncr, aPrimegen, aKeygen, aSign, aSignEncr,
-	   aTest, aPrintMDs, aSignKey,
-    } action = aNull;
     int orig_argc;
     char **orig_argv;
     const char *fname, *fname_print;
@@ -157,9 +189,11 @@ main( int argc, char **argv )
     int default_config =1;
     int errors=0;
     int default_keyring = 1;
+    int greeting = 1;
+    enum cmd_values cmd = 0;
 
 
-    opt.compress = 0; /* defaults to no compression level */
+    opt.compress = -1; /* defaults to standard compress level */
 
     /* check wether we have a config file on the commandline */
     orig_argc = argc;
@@ -213,13 +247,14 @@ main( int argc, char **argv )
 		    break;
 	  case 'z': opt.compress = pargs.r.ret_int; break;
 	  case 'a': opt.armor = 1; opt.no_armor=0; break;
-	  case 'c': action = aSym; break;
+	  case 'c': set_cmd( &cmd , aSym); break;
 	  case 'o': opt.outfile = pargs.r.ret_str; break;
-	  case 'e': action = action == aSign? aSignEncr : aEncr; break;
+	  case 'e': set_cmd( &cmd, aEncr); break;
 	  case 'b': detached_sig = 1;
 	       /* fall trough */
-	  case 's': action = action == aEncr? aSignEncr : aSign;  break;
-	  case 'l': /* store the local users */
+	  case 's': set_cmd( &cmd, aSign );  break;
+	  case 't': set_cmd( &cmd , aClearsig);  break;
+	  case 'u': /* store the local users */
 	    sl = m_alloc( sizeof *sl + strlen(pargs.r.ret_str));
 	    strcpy(sl->d, pargs.r.ret_str);
 	    sl->next = locusr;
@@ -231,21 +266,22 @@ main( int argc, char **argv )
 	    sl->next = remusr;
 	    remusr = sl;
 	    break;
-	  case 500: opt.batch = 1; break;
+	  case 'k': set_cmd( &cmd, aKMode ); break;
+	  case 500: opt.batch = 1; greeting = 0; break;
 	  case 501: opt.answer_yes = 1; break;
 	  case 502: opt.answer_no = 1; break;
-	  case 503: action = aKeygen; break;
-	  case 506: action = aSignKey; break;
-	  case 507: action = aStore; break;
+	  case 503: set_cmd( &cmd, aKeygen); break;
+	  case 506: set_cmd( &cmd, aSignKey); break;
+	  case 507: set_cmd( &cmd, aStore); break;
 	  case 508: opt.check_sigs = 1; opt.list_sigs = 1; break;
 	  case 509: add_keyring(pargs.r.ret_str); nrings++; break;
 	  case 510: opt.debug |= pargs.r.ret_ulong; break;
 	  case 511: opt.debug = ~0; break;
 	  case 512: opt.cache_all = 1; break;
-	  case 513: action = aPrimegen; break;
-	  case 514: action = aTest; break;
+	  case 513: set_cmd( &cmd, aPrimegen); break;
+	  case 514: set_cmd( &cmd, aTest); break;
 	  case 515: opt.fingerprint = 1; break;
-	  case 516: action = aPrintMDs; break;
+	  case 516: set_cmd( &cmd, aPrintMDs); break;
 	  case 517: add_secret_keyring(pargs.r.ret_str); sec_nrings++; break;
 	  case 518:
 	    /* config files may not be nested (silently ignore them) */
@@ -257,6 +293,10 @@ main( int argc, char **argv )
 	    break;
 	  case 519: opt.no_armor=1; opt.armor=0; break;
 	  case 520: default_keyring = 0; break;
+	  case 521: set_cmd( &cmd, aListPackets); break;
+	  case 522: greeting = 0; break;
+	  case 523: set_passphrase_fd( pargs.r.ret_int ); break;
+	  case 524: set_cmd( &cmd, aEditSig); break;
 	  default : errors++; pargs.err = configfp? 1:2; break;
 	}
     }
@@ -271,13 +311,26 @@ main( int argc, char **argv )
 	exit(2);
 
     set_debug();
+    if( cmd == aKMode || cmd == aKModeC ) { /* kludge to be compatible to pgp */
+	if( cmd == aKModeC ) {
+	    opt.fingerprint = 1;
+	    cmd = aKMode;
+	}
+	opt.list_sigs = 0;
+	if( opt.verbose > 2 )
+	    opt.check_sigs++;
+	if( opt.verbose > 1 )
+	    opt.list_sigs++;
+
+	opt.verbose = opt.verbose > 1;
+    }
     if( opt.verbose > 1 )
 	set_packet_list_mode(1);
-    if( opt.verbose && isatty(fileno(stdin)) ) {
+    if( greeting ) {
 	if( *(s=strusage(10))  )
-	    fputs(s, stderr);
+	    tty_printf("%s", s);
 	if( *(s=strusage(30))  )
-	    fputs(s, stderr);
+	    tty_printf("%s", s);
     }
 
     if( !sec_nrings || default_keyring ) { /* add default secret rings */
@@ -299,7 +352,7 @@ main( int argc, char **argv )
 	fname = NULL;
     }
 
-    switch( action ) {
+    switch( cmd ) {
       case aStore: /* only store the file */
 	if( argc > 1 )
 	    usage(1);
@@ -325,14 +378,15 @@ main( int argc, char **argv )
       case aSign: /* sign the given file */
 	if( argc > 1 )
 	    usage(1);
-	if( (rc = sign_file(fname, detached_sig, locusr)) )
+	if( (rc = sign_file(fname, detached_sig, locusr, 0, NULL)) )
 	    log_error("sign_file('%s'): %s\n", fname_print, g10_errstr(rc) );
 	break;
 
-
       case aSignEncr: /* sign and encrypt the given file */
-	log_fatal("signing and encryption is not yet implemented\n");
-	usage(1);  /* FIXME */
+	if( argc > 1 )
+	    usage(1);
+	if( (rc = sign_file(fname, detached_sig, locusr, 1, remusr)) )
+	    log_error("sign_file('%s'): %s\n", fname_print, g10_errstr(rc) );
 	break;
 
 
@@ -344,6 +398,45 @@ main( int argc, char **argv )
 	    log_error("sign_key('%s'): %s\n", fname_print, g10_errstr(rc) );
 	break;
 
+      case aEditSig: /* Edit a key signature */
+	if( argc != 1 )
+	    usage(1);
+	/* note: fname is the user id! */
+	if( (rc = edit_keysigs(fname)) )
+	    log_error("edit_keysig('%s'): %s\n", fname_print, g10_errstr(rc) );
+	break;
+
+      case aKMode: /* list keyring */
+	if( !argc ) { /* list the default public keyrings */
+	    int i, seq=0;
+	    const char *s;
+
+	    while( s=get_keyring(seq++) ) {
+		if( !(a = iobuf_open(s)) ) {
+		    log_error("can't open '%s'\n", s);
+		    continue;
+		}
+		if( seq > 1 )
+		    putchar('\n');
+		printf("%s\n", s );
+		for(i=strlen(s); i; i-- )
+		    putchar('-');
+		putchar('\n');
+
+		proc_packets( a );
+		iobuf_close(a);
+	    }
+
+	}
+	else if( argc == 1) { /* list the given keyring */
+	    if( !(a = iobuf_open(fname)) )
+		log_fatal("can't open '%s'\n", fname_print);
+	    proc_packets( a );
+	    iobuf_close(a);
+	}
+	else
+	    usage(1);
+	break;
 
       case aPrimegen:
 	if( argc == 1 ) {
@@ -385,6 +478,8 @@ main( int argc, char **argv )
 
       case aTest: do_test( argc? atoi(*argv): 0 ); break;
 
+      case aListPackets:
+	opt.list_packets=1;
       default:
 	if( argc > 1 )
 	    usage(1);
@@ -395,6 +490,10 @@ main( int argc, char **argv )
 	    memset( &afx, 0, sizeof afx);
 	    iobuf_push_filter( a, armor_filter, &afx );
 	}
+	if( cmd == aListPackets ) {
+	    set_packet_list_mode(1);
+	    opt.list_packets=1;
+	}
 	proc_packets( a );
 	iobuf_close(a);
 	break;
@@ -403,7 +502,7 @@ main( int argc, char **argv )
     /* cleanup */
     FREE_STRLIST(remusr);
     FREE_STRLIST(locusr);
-    return 0;
+    return log_get_errorcount(0)? 2:0;
 }
 
 
@@ -506,33 +605,5 @@ do_test(int times)
 
     m_check(NULL);
   #endif
-  #if 0
-    char *array;
-    int i, j;
-    int n = 6;
-    int m = times;
-
-    if( m > n )
-	abort();
-    array = m_alloc_clear( n );
-    memset( array, 1, m );
-
-    for(i=0;; i++) {
-	printf("i=%3d: ", i );
-	for(j=0; j < n ; j++ )
-	    if( array[j] )
-		putchar( 'X' );
-	    else
-		putchar( '-' );
-	putchar('\n');
-	m_out_of_n( array, m, n );
-	for(j=0; j < n; j++ )
-	    if( !array[j] )
-		break;
-	if( j == m )
-	    break;
-    }
-  #endif
 }
 
-
