@@ -37,92 +37,6 @@
 #include "ttyio.h"
 
 
-typedef struct skc_list *SKC_LIST;
-struct skc_list {
-    SKC_LIST next;
-    PKT_secret_cert *skc;
-    int mark;
-};
-
-
-static void
-release_skc_list( SKC_LIST skc_list )
-{
-    SKC_LIST skc_rover;
-
-    for( ; skc_list; skc_list = skc_rover ) {
-	skc_rover = skc_list->next;
-	free_secret_cert( skc_list->skc );
-	m_free( skc_list );
-    }
-}
-
-static int
-build_skc_list( STRLIST locusr, SKC_LIST *ret_skc_list, int unlock )
-{
-    SKC_LIST skc_list = NULL;
-    SKC_LIST skc_rover = NULL;
-    int rc;
-
-    if( !locusr ) { /* use the default one */
-	PKT_secret_cert *skc;
-
-	skc = m_alloc_clear( sizeof *skc );
-	if( (rc = get_seckey_byname( skc, NULL, unlock )) ) {
-	    free_secret_cert( skc ); skc = NULL;
-	    log_error("no default secret key: %s\n", g10_errstr(rc) );
-	}
-	else if( is_valid_pubkey_algo(skc->pubkey_algo) ) {
-	    SKC_LIST r;
-	    r = m_alloc( sizeof *r );
-	    r->skc = skc; skc = NULL;
-	    r->next = skc_list;
-	    r->mark = 0;
-	    skc_list = r;
-	}
-	else {
-	    free_secret_cert( skc ); skc = NULL;
-	    log_error("invalid default secret key: %s\n", g10_errstr(rc) );
-	}
-    }
-    else {
-	for(; locusr; locusr = locusr->next ) {
-	    PKT_secret_cert *skc;
-
-	    skc = m_alloc_clear( sizeof *skc );
-	    if( (rc = get_seckey_byname( skc, locusr->d, unlock )) ) {
-		free_secret_cert( skc ); skc = NULL;
-		log_error("skipped '%s': %s\n", locusr->d, g10_errstr(rc) );
-	    }
-	    else if ( is_valid_pubkey_algo(skc->pubkey_algo) ) {
-		SKC_LIST r;
-		r = m_alloc( sizeof *r );
-		r->skc = skc; skc = NULL;
-		r->next = skc_list;
-		r->mark = 0;
-		skc_list = r;
-	    }
-	    else {
-		free_secret_cert( skc ); skc = NULL;
-		log_error("skipped '%s': %s\n", locusr->d, g10_errstr(rc) );
-	    }
-	}
-    }
-
-
-    if( !rc && !skc_list ) {
-	log_error("no valid signators\n");
-	rc = G10ERR_NO_USER_ID;
-    }
-
-    if( rc )
-	release_skc_list( skc_list );
-    else
-	*ret_skc_list = skc_list;
-    return rc;
-}
-
-
 
 
 
@@ -130,61 +44,13 @@ static int
 complete_sig( PKT_signature *sig, PKT_secret_cert *skc, MD_HANDLE *md )
 {
     int rc=0;
-    byte *dp;
-
-    dp = md_final( md );
 
     if( (rc=check_secret_key( skc )) )
 	;
-    else if( sig->pubkey_algo == PUBKEY_ALGO_ELGAMAL ) {
-	ELG_secret_key skey;
-	MPI frame;
-
-	keyid_from_skc( skc, sig->keyid );
-	sig->d.elg.digest_algo = DIGEST_ALGO_RMD160;
-	sig->d.elg.digest_start[0] = dp[0];
-	sig->d.elg.digest_start[1] = dp[1];
-	sig->d.elg.a = mpi_alloc( mpi_get_nlimbs(skc->d.elg.p) );
-	sig->d.elg.b = mpi_alloc( mpi_get_nlimbs(skc->d.elg.p) );
-	frame = encode_md_value( md, mpi_get_nbits(skc->d.elg.p));
-	skey.p = skc->d.elg.p;
-	skey.g = skc->d.elg.g;
-	skey.y = skc->d.elg.y;
-	skey.x = skc->d.elg.x;
-	elg_sign( sig->d.elg.a, sig->d.elg.b, frame, &skey);
-	memset( &skey, 0, sizeof skey );
-	mpi_free(frame);
-	if( opt.verbose ) {
-	    char *ustr = get_user_id_string( sig->keyid );
-	    log_info("ElGamal signature from: %s\n", ustr );
-	    m_free(ustr);
-	}
-    }
-  #ifdef HAVE_RSA_CIPHER
-    else if( sig->pubkey_algo == PUBKEY_ALGO_RSA ) {
-	RSA_secret_key skey;
-
-	keyid_from_skc( skc, sig->keyid );
-	sig->d.rsa.digest_algo = DIGEST_ALGO_RMD160;
-	sig->d.rsa.digest_start[0] = dp[0];
-	sig->d.rsa.digest_start[1] = dp[1];
-	sig->d.rsa.rsa_integer =
-		       encode_md_value( md, mpi_get_nbits(skc->d.rsa.rsa_n));
-	skey.e = skc->d.rsa.rsa_e;
-	skey.n = skc->d.rsa.rsa_n;
-	skey.p = skc->d.rsa.rsa_p;
-	skey.q = skc->d.rsa.rsa_q;
-	skey.d = skc->d.rsa.rsa_d;
-	skey.u = skc->d.rsa.rsa_u;
-	rsa_secret( sig->d.rsa.rsa_integer, sig->d.rsa.rsa_integer, &skey);
-	memset( &skey, 0, sizeof skey );
-	if( opt.verbose ) {
-	    char *ustr = get_user_id_string( sig->keyid );
-	    log_info("RSA signature from: %s\n", ustr );
-	    m_free(ustr);
-	}
-    }
-  #endif/*HAVE_RSA_CIPHER*/
+    else if( sig->pubkey_algo == PUBKEY_ALGO_ELGAMAL )
+	g10_elg_sign( skc, sig, md );
+    else if( sig->pubkey_algo == PUBKEY_ALGO_RSA )
+	g10_rsa_sign( skc, sig, md );
     else
 	log_bug(NULL);
 
@@ -202,28 +68,40 @@ complete_sig( PKT_signature *sig, PKT_secret_cert *skc, MD_HANDLE *md )
  * make a detached signature.  If FILENAME is NULL read from stdin
  * and ignore the detached mode.  Sign the file with all secret keys
  * which can be taken from LOCUSR, if this is NULL, use the default one
+ * If ENCRYPT is true, use REMUSER (or ask if it is NULL) to encrypt the
+ * signed data for these users.
  */
 int
-sign_file( const char *filename, int detached, STRLIST locusr )
+sign_file( const char *filename, int detached, STRLIST locusr,
+	   int encrypt, STRLIST remusr )
 {
     armor_filter_context_t afx;
     compress_filter_context_t zfx;
     md_filter_context_t mfx;
+    text_filter_context_t tfx;
+    encrypt_filter_context_t efx;
     IOBUF inp = NULL, out = NULL;
     PACKET pkt;
     PKT_plaintext *pt = NULL;
     u32 filesize;
     int last_rc, rc = 0;
+    PKC_LIST pkc_list = NULL;
     SKC_LIST skc_list = NULL;
     SKC_LIST skc_rover = NULL;
 
     memset( &afx, 0, sizeof afx);
     memset( &zfx, 0, sizeof zfx);
     memset( &mfx, 0, sizeof mfx);
+    memset( &tfx, 0, sizeof tfx);
+    memset( &efx, 0, sizeof efx);
     init_packet( &pkt );
 
     if( (rc=build_skc_list( locusr, &skc_list, 1 )) )
 	goto leave;
+    if( encrypt ) {
+	if( (rc=build_pkc_list( remusr, &pkc_list )) )
+	    goto leave;
+    }
 
     /* prepare iobufs */
     if( !(inp = iobuf_open(filename)) ) {
@@ -233,12 +111,14 @@ sign_file( const char *filename, int detached, STRLIST locusr )
 	goto leave;
     }
 
-    if( !(out = open_outfile( filename )) ) {
+    if( !(out = open_outfile( filename, opt.armor? 1: detached? 2:0 )) ) {
 	rc = G10ERR_CREATE_FILE;
 	goto leave;
     }
 
     /* prepare to calculate the MD over the input */
+    if( opt.textmode && opt.armor )
+	iobuf_push_filter( inp, text_filter, &tfx );
     mfx.rmd160 = rmd160_open(0);
     iobuf_push_filter( inp, md_filter, &mfx );
 
@@ -248,6 +128,11 @@ sign_file( const char *filename, int detached, STRLIST locusr )
     if( opt.compress )
 	iobuf_push_filter( out, compress_filter, &zfx );
 
+    if( encrypt ) {
+	efx.pkc_list = pkc_list;
+	/* fixme: set efx.cfx.datalen if known */
+	iobuf_push_filter( out, encrypt_filter, &efx );
+    }
 
     /* loop over the secret certificates and build headers */
     for( skc_rover = skc_list; skc_rover; skc_rover = skc_rover->next ) {
@@ -256,7 +141,7 @@ sign_file( const char *filename, int detached, STRLIST locusr )
 
 	skc = skc_rover->skc;
 	ops = m_alloc_clear( sizeof *ops );
-	ops->sig_class = 0x00;
+	ops->sig_class = opt.textmode? 0x01 : 0x00;
 	ops->digest_algo = DIGEST_ALGO_RMD160;
 	ops->pubkey_algo = skc->pubkey_algo;
 	keyid_from_skc( skc, ops->keyid );
@@ -276,7 +161,7 @@ sign_file( const char *filename, int detached, STRLIST locusr )
 
     /* setup the inner packet */
     if( detached ) {
-	/* read, so that the filter can calculate the digests */
+	/* read, so that the filter can calculate the digest */
 	while( iobuf_get(inp) != -1 )
 	    ;
     }
@@ -294,7 +179,7 @@ sign_file( const char *filename, int detached, STRLIST locusr )
 	    filesize = 0; /* stdin */
 	}
 	pt->timestamp = make_timestamp();
-	pt->mode = 'b';
+	pt->mode = opt.textmode? 't':'b';
 	pt->len = filesize;
 	pt->buf = inp;
 	pkt.pkttype = PKT_PLAINTEXT;
@@ -318,7 +203,7 @@ sign_file( const char *filename, int detached, STRLIST locusr )
 	sig = m_alloc_clear( sizeof *sig );
 	sig->pubkey_algo = skc->pubkey_algo;
 	sig->timestamp = make_timestamp();
-	sig->sig_class = 0x00;
+	sig->sig_class = opt.textmode? 0x01 : 0x00;
 
 	rmd = rmd160_copy( mfx.rmd160 );
 	rmd160_putchar( rmd, sig->sig_class );
@@ -399,13 +284,14 @@ sign_file( const char *filename, int detached, STRLIST locusr )
 
 
   leave:
-    release_skc_list( skc_list );
     if( rc )
 	iobuf_cancel(out);
     else
 	iobuf_close(out);
     iobuf_close(inp);
     rmd160_close( mfx.rmd160 );
+    release_skc_list( skc_list );
+    release_pkc_list( pkc_list );
     return rc;
 }
 
@@ -465,7 +351,15 @@ sign_it_p( PKT_public_cert *pkc, PKT_user_id *uid )
 }
 
 
-static void
+/****************
+ * Check the keysigs and set the flags to indicate errors.
+ * Usage of nodes flag bits:
+ * Bit	0 = bad signature
+ *	1 = no public key
+ *	2 = other error
+ * Returns true if error found.
+ */
+static int
 check_all_keysigs( KBNODE keyblock )
 {
     KBNODE kbctx;
@@ -501,6 +395,7 @@ check_all_keysigs( KBNODE keyblock )
 		m_free(p);
 	    }
 	    tty_printf("\n");
+	    /* FIXME: update the trustdb */
 	}
     }
     if( inv_sigs )
@@ -509,6 +404,76 @@ check_all_keysigs( KBNODE keyblock )
 	tty_printf("No public key for %d signatures\n", no_key );
     if( oth_err )
 	tty_printf("%d signatures not checked due to errors\n", oth_err );
+    return inv_sigs || no_key || oth_err;
+}
+
+
+/****************
+ * Ask and remove invalid signatures are to be removed.
+ */
+static int
+remove_keysigs( KBNODE keyblock, int all )
+{
+    KBNODE kbctx;
+    KBNODE node;
+    char *answer;
+    int yes;
+    int count;
+
+    count = 0;
+    for( kbctx=NULL; (node=walk_kbtree( keyblock, &kbctx)) ; ) {
+	if( ((node->flag & 7) || all )
+	    && node->pkt->pkttype == PKT_SIGNATURE
+	    && (node->pkt->pkt.signature->sig_class&~3) == 0x10 ) {
+	    PKT_signature *sig = node->pkt->pkt.signature;
+	    int sigrc;
+
+	    if( all ) {
+		/* fixme: skip self-sig */
+	    }
+
+	    tty_printf("\n \"%08lX %s   ",
+			sig->keyid[1], datestr_from_sig(sig));
+	    if( node->flag & 6 )
+		tty_printf("[User name not available] ");
+	    else {
+		size_t n;
+		char *p = get_user_id( sig->keyid, &n );
+		tty_print_string( p, n );
+		m_free(p);
+	    }
+	    tty_printf("\"\n");
+	    if( node->flag & 1 )
+		tty_printf("This is a BAD signature!\n");
+	    else if( node->flag & 2 )
+		tty_printf("Public key not available.\n");
+	    else if( node->flag & 4 )
+		tty_printf("The signature could not be checked!\n");
+	    answer = tty_get("\nRemove this signature? ");
+	    tty_kill_prompt();
+	    if( answer_is_yes(answer) ) {
+		node->flag |= 128;     /* use bit 7 to mark this node */
+		count++;
+	    }
+	    m_free(answer);
+	}
+    }
+
+    if( !count )
+	return 0; /* nothing to remove */
+    answer = tty_get("Do you really want to remove the selected signatures? ");
+    tty_kill_prompt();
+    yes = answer_is_yes(answer);
+    m_free(answer);
+    if( !yes )
+	return 0;
+
+    for( kbctx=NULL; (node=walk_kbtree2( keyblock, &kbctx, 1)) ; ) {
+	if( node->flag & 128)
+	    delete_kbnode( keyblock, node );
+    }
+
+    return 1;
 }
 
 
@@ -531,6 +496,7 @@ sign_key( const char *username, STRLIST locusr )
     PKT_public_cert *pkc;
     int any;
     u32 pkc_keyid[2];
+    char *answer;
 
     memset( &mfx, 0, sizeof mfx);
 
@@ -581,9 +547,16 @@ sign_key( const char *username, STRLIST locusr )
     }
 
     clear_kbnode_flags( keyblock );
-    check_all_keysigs( keyblock );
-    /* look wether we should ask to remove invalid keys */
-    /*+ FIXME: */
+    if( check_all_keysigs( keyblock ) ) {
+	if( !opt.batch ) {
+	    /* ask wether we really should do anything */
+	    answer = tty_get("To you want to remove some of the invalid sigs? ");
+	    tty_kill_prompt();
+	    if( answer_is_yes(answer) )
+		remove_keysigs( keyblock, 0 );
+	    m_free(answer);
+	}
+    }
 
     /* check wether we have already signed it */
     for( skc_rover = skc_list; skc_rover; skc_rover = skc_rover->next ) {
@@ -641,7 +614,7 @@ sign_key( const char *username, STRLIST locusr )
 
     rc = update_keyblock( &kbpos, keyblock );
     if( rc ) {
-	log_error("insert_keyblock failed: %s\n", g10_errstr(rc) );
+	log_error("update_keyblock failed: %s\n", g10_errstr(rc) );
 	goto leave;
     }
 
@@ -654,6 +627,75 @@ sign_key( const char *username, STRLIST locusr )
 
 
 
+int
+edit_keysigs( const char *username )
+{
+    int rc = 0;
+    KBNODE keyblock = NULL;
+    KBNODE kbctx, node;
+    KBPOS kbpos;
+    PKT_public_cert *pkc;
+    int any;
+    u32 pkc_keyid[2];
+    char *answer;
+
+    /* search the userid */
+    rc = search_keyblock_byname( &kbpos, username );
+    if( rc ) {
+	log_error("user '%s' not found\n", username );
+	goto leave;
+    }
+
+    /* read the keyblock */
+    rc = read_keyblock( &kbpos, &keyblock );
+    if( rc ) {
+	log_error("error reading the certificate: %s\n", g10_errstr(rc) );
+	goto leave;
+    }
+
+    /* get the keyid from the keyblock */
+    for( kbctx=NULL; (node=walk_kbtree( keyblock, &kbctx)) ; ) {
+	if( node->pkt->pkttype == PKT_PUBLIC_CERT )
+	    break;
+    }
+    if( !node ) {
+	log_error("Oops; public key not found anymore!\n");
+	rc = G10ERR_GENERAL;
+	goto leave;
+    }
+
+    pkc = node->pkt->pkt.public_cert;
+    keyid_from_pkc( pkc, pkc_keyid );
+    log_info("Checking signatures of this public key certificate:\n");
+    tty_printf("pub  %4u%c/%08lX %s   ",
+	      nbits_from_pkc( pkc ),
+	      pubkey_letter( pkc->pubkey_algo ),
+	      pkc_keyid[1], datestr_from_pkc(pkc) );
+    {
+	size_t n;
+	char *p = get_user_id( pkc_keyid, &n );
+	tty_print_string( p, n > 40? 40 : n );
+	m_free(p);
+	tty_printf("\n");
+    }
+
+    clear_kbnode_flags( keyblock );
+    check_all_keysigs( keyblock );
+    if( remove_keysigs( keyblock, 1 ) ) {
+	rc = update_keyblock( &kbpos, keyblock );
+	if( rc ) {
+	    log_error("update_keyblock failed: %s\n", g10_errstr(rc) );
+	    goto leave;
+	}
+    }
+
+  leave:
+    release_kbnode( keyblock );
+    return rc;
+}
+
+
+
 /****************
  * Create a signature packet for the given public key certificate
  * and the user id and return it in ret_sig. User signature class SIGCLASS
